;; Generated package description from logos.el  -*- no-byte-compile: t -*-
(define-package "logos" "0.4.0" "Simple focus mode and extras" '((emacs "27.1")) :commit "388bfbd9d8c880898d1832e888d5b41582dd1a85" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "focus" "writing" "presentation" "narrowing") :url "https://git.sr.ht/~protesilaos/logos")
